/*==============================================================================
** Copyright (c)  2002 - 2007 Xerox Corporation. All Rights Reserved.
** Copyright protection claimed includes all forms and matters of
** copyrightable material and information now allowed by statutory or
** judicial law or hereinafter granted, including without limitation,
** material generated from the software programs which are displayed
** on the screen such as icons, screen and the like.
*/
/*==============================================================================
**
** DESCRIPTION: 
**     This module implements the Samba-to-CPI interface
** REVISION HISTORY:
**
** AUTHOR        DATE       DESCRIPTION 
** H. Wang       07/03/08   CQGbl82973: Upgrade to Samba 3.0.30
** M. Lucyshyn   07/12/07   CDDub23022: Upgrade to Samba 3.0.25b
** M. Lucyshyn   08/30/06   CDDub13605: Upgrade to Samba 3.0.22
** M. Lucyshyn   03/31/05   CDDua87041: Upgrade to Samba 3.0.13
** M. Lucyshyn   02/04/04   CDDua68494: Upgrade to smbclient 3.0.1
** M. Lucyshyn   03/08/15   CDDua64508: Job Size and job identifier added to CPI
**                          print queue information.
** M. Lucyshyn   02/09/25   CDDua49978: Continue integration of CPI interface 
**                          into Samba to support printing.
** T. Moody      02/08/06   CDDua49263 Integrate use of CPI interface into Samba
**                          to support printing.
** T. Moody      02/07/15   Created
*******************************************************************************/

/*   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef XEROX_MODIFICATION

#include "includes.h"
#include "printing.h"


#define XLOGENTER  DEBUG(3,("Enter: %s()\n", __FUNCTION__));
#define XLOGEXIT   DEBUG(3,("Exit: %s():%d\n", __FUNCTION__, __LINE__));


/*
 * Generic printing interface definitions...
 */


static int cpi_job_delete(const char *sharename, const char *lprm_command,
                           struct printjob * pjob);
static int cpi_job_pause(int snum, struct printjob *pjob);
static int cpi_job_resume(int snum, struct printjob *pjob);
static int cpi_job_submit(int snum, struct printjob *pjob);
static int cpi_queue_get(const char *printer_name,
                   enum printing_types printing_type,
                   char *lpq_command,
                   print_queue_struct **q,
                   print_status_struct *status);
static int cpi_queue_pause(int snum);
static int cpi_queue_resume(int snum);


struct printif  cpi_printif =
{
    PRINT_BSD,
    cpi_queue_get,
    cpi_queue_pause,
    cpi_queue_resume,
    cpi_job_delete,
    cpi_job_pause,
    cpi_job_resume,
    cpi_job_submit,
};

extern int DEBUGLEVEL;

/****************************************************************************
delete a print job
****************************************************************************/
static int cpi_job_delete (const char *sharename, const char *lprm_command,
                           struct printjob * pjob)
{
    cpi_status_t cpi_status;
    int retVal = 0;

    XLOGENTER;
    
    cpi_status = CPI_Job_Cancel( pjob->pcpi_client_handle,
                                 CPI_JOBH_ID_P(pjob->cpi_pjobh) );
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("CPI_Job_Cancel() failed at func %s():%d; "
                 "code= %d for CPI job %d\n",
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status),
                 CPI_JOBH_ID_P(pjob->cpi_pjobh)));
        retVal = -1;
    }

    XLOGEXIT;
    return( retVal );
}

/****************************************************************************
pause a job
****************************************************************************/
static int cpi_job_pause(int snum, struct printjob *pjob)
{
    cpi_status_t cpi_status;
    int retVal = 0;

    XLOGENTER;

    cpi_status = CPI_Job_Pause( pjob->pcpi_client_handle,
                                CPI_JOBH_ID_P(pjob->cpi_pjobh) );
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("CPI_Job_Pause() failed at func %s():%d; "
                 "code= %d for CPI job %d\n",
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status),
                 CPI_JOBH_ID_P(pjob->cpi_pjobh)));
        retVal = -1;
    }

    XLOGEXIT;
    return( retVal );
}

/****************************************************************************
resume a job
****************************************************************************/
static int cpi_job_resume(int snum, struct printjob *pjob)
{
    cpi_status_t cpi_status;
    int          retVal = 0;

    XLOGENTER;

    cpi_status = CPI_Job_Resume( pjob->pcpi_client_handle,
                                 CPI_JOBH_ID_P(pjob->cpi_pjobh) );
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("CPI_Job_Pause() failed at func %s():%d; "
                 "code= %d for CPI job %d\n", 
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status),
                 CPI_JOBH_ID_P(pjob->cpi_pjobh)));
        retVal = -1;
    }

    XLOGEXIT;
    return( retVal );
}

/****************************************************************************
 Submit a file for printing - called from print_job_end()
****************************************************************************/
/* Samba expects to spool the entire job to file, then submit the file.  With
   CPI, there is no temporary spooling to file; instead the fd written to is
   that of a pipe to the DCH backend.  Therefore, when all data is "written",
   it is fully submitted, so the Samba 'job submit' function is where the
   job is completed for CPI. */
static int cpi_job_submit(int snum, struct printjob *pjob)
{
    cpi_status_t cpi_status;

    XLOGENTER;

    cpi_status = CPI_Doc_Complete(pjob->pcpi_client_handle, pjob->cpi_pjobh,
                                  CPI_COMPL_NO_ANSWER, NULL /*cpi doc status*/);
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("%s():%d: CPI_Doc_Complete() failed at func %s():%d; "
                 "code= %d\n", __FUNCTION__, __LINE__,
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status)));
        /* No point in returning, keep trying to finish the job */
    }

    cpi_status = CPI_Job_Complete(pjob->pcpi_client_handle, &pjob->cpi_pjobh,
                    CPI_COMPL_NO_ANSWER, NULL/*cpi job status*/);
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("%s():%d: Failed at func %s():%d; "
                 "code= %d \n", __FUNCTION__, __LINE__,
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status)));
    
    }

    cpi_status = CPI_Client_Shutdown( &pjob->pcpi_client_handle,
                                      True /*wait_for_answer*/ );
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("%s():%d: CPI_Client_Shutdown() failed at func %s():%d; "
                 "code= %d\n", __FUNCTION__, __LINE__,
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status)));
        XLOGEXIT;
        return( -1 );
    }

    XLOGEXIT;
    return (0);
}


/****************************************************************************
get the current list of queued jobs
****************************************************************************/
static int cpi_queue_get(const char *printer_name, 
                        enum printing_types printing_type,
                        char *lpq_command,
                        print_queue_struct **q,
                        print_status_struct *status)

{
    cpi_status_t cpi_status;
    cpi_client_handle_t  *ph = NULL;
    cpi_queue_info_handle_t  *pqh = NULL;
    int  job_count = 0, i;
    cpi_job_queue_info_t jobinfo;
    cpi_bool_t     retValid;
    struct tm  time_tm;
    char       timestr[128]; 
    print_queue_struct *pqueue = NULL;
    print_queue_struct *pqueue_job = NULL;
    struct printjob * pJob;
    size_t udataSize;

    XLOGENTER;
    
    CPI_SET_STATUS( cpi_status, CPI_STATUS_OK ); /* init */

    if( q == NULL )
    {
        DEBUG(3,("%s():%d: q is NULL!\n", __FUNCTION__, __LINE__));
        return( -1 );
    }
    
    /* Establish a new CPI client for this transaction */
    cpi_status = CPI_Client_Init( "SMBD", 30 /*TIMEOUT_IN_SEC*/, &ph );
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("%s():%d: CPI_Client_Init() failed at func %s():%d; "
                 "code= %d\n", __FUNCTION__, __LINE__,
                 CPI_GET_STATUS_FUNC(cpi_status),
                 CPI_GET_STATUS_LINE(cpi_status),
                 CPI_GET_STATUS_CODE(cpi_status)));
        CPI_Client_Shutdown( &ph, True /*wait_for_answer*/ );
        XLOGEXIT;
        return( -1 );
    }
    
    cpi_status = CPI_Get_Job_Queue_Info( ph, &pqh );
    if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
    {
        DEBUG(3,("CPI_Get_Job_Queue_Info() failed\n"));
        CPI_Client_Shutdown( &ph, True /*wait_for_answer*/ );
        XLOGEXIT;
        return( -1 );
    }

    DEBUG(3,("Number of jobs in queue from print_queue_handle = %d\n",
            pqh->num_jobs_in_queue_info));
    job_count = 0;
    for( i = 0; i < pqh->num_jobs_in_queue_info; i++)
    {
        cpi_status = CPI_Get_Job_Queue_Record( pqh, (job_count+1), &jobinfo,
                                               &retValid );
        if( CPI_GET_STATUS_CODE(cpi_status) != CPI_STATUS_OK )
        {
            DEBUG(3,("CPI_Get_Job_Queue_Record() failed\n"));
            break;
        }

        if( retValid == False )
        {
            if( job_count == 0 )
            {
                DEBUG(3,("No jobs\n"));
            }
            break;
        }
        else
        {
            job_count++;
            memcpy( &time_tm, localtime(&jobinfo.time), sizeof(struct tm));
            strftime( timestr, 128, "%B %d, %Y   %H:%M:%S", &time_tm);
            DEBUG(3,("CPI queue information received:\n\tJob Number: %3d:\n"
                     "\tJob id:      %d\n\tJob Status: %s"
                     "\n\tTime:        %s\n\tJob Owner:   %s\n\tJob Name:    "
                     "%s\n\tJob Idenifier: %s\n\tSize:          %d\n"
                     "\tUser Data:     %s\n", 
                     job_count,
                     jobinfo.job_id,
                     cpi_job_status_string(jobinfo.job_status),
                     timestr,
                     jobinfo.job_owner,
                     jobinfo.job_name,
                     jobinfo.job_identifier,
                     jobinfo.job_size,
                     jobinfo.udata.data ));

            pqueue = (print_queue_struct *)realloc(pqueue,
                                  (job_count * sizeof(print_queue_struct)));
            if( pqueue == NULL )
            {
                DEBUG(3,("Failed to realloc space for Samba queue "
                         "entry!\n"));
                DEBUG(3,("Job count = %d\n", job_count));
                break;
            }

            pqueue_job = &pqueue[job_count-1];
            pqueue_job->job = jobinfo.job_id;

			/*
			** Copy over the jobname into the fs_file member.
			*/
			
			fstrcpy(pqueue_job->fs_file, jobinfo.job_name);
			
            switch( jobinfo.job_status )
            {
                case CPI_JOB_STATUS_QUEUED:
                    pqueue_job->status = LPQ_QUEUED;
                    break;
                case CPI_JOB_STATUS_PAUSED:
                    pqueue_job->status = LPQ_PAUSED;
                    break;
                case CPI_JOB_STATUS_SPOOLING:
                    pqueue_job->status = LPQ_SPOOLING;
                    break;
                case CPI_JOB_STATUS_PRINTING:
                    pqueue_job->status = LPQ_PRINTING;
                    break;
                case CPI_JOB_STATUS_ERROR:
                    pqueue_job->status = LPQ_ERROR;
                    break;
                case CPI_JOB_STATUS_DELETING:
                    pqueue_job->status = LPQ_DELETING; 
                    break;
                case CPI_JOB_STATUS_OFFLINE:
                    pqueue_job->status = LPQ_OFFLINE; 
                    break;
                case CPI_JOB_STATUS_PRINTED:
                    pqueue_job->status = LPQ_PRINTED; 
                    break;                
                case CPI_JOB_STATUS_CANCELLED:
                    pqueue_job->status = LPQ_DELETED; 
                    break;                
                case CPI_JOB_STATUS_BLOCKED:
                    pqueue_job->status = LPQ_BLOCKED;
                    break;                
                
                default:
                    DEBUG(3,("CPI job state = %d not recognized\n",
                             jobinfo.job_status));
                    pqueue_job->status = LPQ_ERROR;
                    break;                                    

                   
            }
            pqueue_job->priority = 1;
            pqueue_job->time = jobinfo.time;
            fstrcpy( pqueue_job->fs_user, jobinfo.job_owner); 
            pqueue_job->size = jobinfo.job_size;
            pqueue_job->page_count = 0; /* Not supported yet */

            DEBUG(3,("The contents of the print_queue structure that will be"
                  " returned:\n\tjob:  %d\n\tsize: %d\n\tpage_count: %d\n"
                  "\tstatus: %d\n\tpriority %d\n\ttime %s\n\tuser: %s\n\tfile:"
                  " %s\n",
                  pqueue_job->job, pqueue_job->size, pqueue_job->page_count, 
                  pqueue_job->status, pqueue_job->priority, timestr,
                  pqueue_job->fs_user, pqueue_job->fs_file ));
        }
    } /* end while */
    CPI_Free_Job_Queue_Info_Handle( &pqh );        

    CPI_Client_Shutdown( &ph, True /*wait_for_answer*/ );

    if( pqueue != NULL )
    {
        *q = pqueue;
        XLOGEXIT;
        return( job_count );
    }
    else
    {
        XLOGEXIT;
        return( 0 );
    }
}

/****************************************************************************
 pause a queue
****************************************************************************/
static int cpi_queue_pause(int snum)
{
    XLOGENTER;
    XLOGEXIT;
    return( 1 );
}

/****************************************************************************
 resume a queue
****************************************************************************/
static int cpi_queue_resume(int snum)
{
    XLOGENTER;
    XLOGEXIT;
    return( 1 );
}


#endif /* XEROX_MODIFICATION */
